@@ -40,6 +40,7 @@
 #endif
 
 #include <stdio.h>
+#include <math.h>
 #include "MaemoLocationProvider.h"
 #include "nsIClassInfo.h"
 #include "nsDOMClassInfoID.h"
@@ -55,8 +56,8 @@
 #include "MaemoLocationPrompt.h"
 #endif
 
-//#define DEBUG 1
-#ifdef DEBUG
+//#define LOGGING 1
+#ifdef LOGGING
 #define OSSOLOG_COMPILE 1
 #include "osso-log.h"
 #endif
@@ -207,7 +208,11 @@ MaemoLocationProvider::MaemoLocationProvider() :
   mHasSeenLocation(PR_FALSE),
   mHasGPS(PR_TRUE),
   mGPSControl(nsnull),
-  mGPSDevice(nsnull)
+  mGPSDevice(nsnull),
+  mIgnoreMinorChanges(PR_FALSE),
+  mPrevLat(0.0),
+  mPrevLong(0.0),
+  mInterval(LOCATION_INTERVAL_5S)
 {
 }
 
@@ -231,8 +236,23 @@ void MaemoLocationProvider::ControlError(LocationGPSDControl* control, void* sel
 
 void MaemoLocationProvider::LocationChanged(LocationGPSDevice* device, void* self)
 {
-  NS_ENSURE_TRUE(self, );
+  if (!device || !device->fix)
+    return;
+
+  guint32 &fields = device->fix->fields;
+  if (!(fields & LOCATION_GPS_DEVICE_LATLONG_SET))
+    return;
+  if (!(device->fix->eph && !isnan(device->fix->eph)))
+    return;
+
   MaemoLocationProvider* provider = static_cast<MaemoLocationProvider*>(self);
+  NS_ENSURE_TRUE(provider, );
+  provider->LocationUpdate(device);
+}
+
+nsresult
+MaemoLocationProvider::LocationUpdate(LocationGPSDevice* aDev)
+{
 #ifdef TEST_BUILD_WITHOUT_DEVICE
   nsRefPtr<nsGeoPosition> somewhere = new nsGeoPosition(60.207075, // Lat
                                                         25.125389, // Long
@@ -243,33 +263,41 @@ void MaemoLocationProvider::LocationChanged(LocationGPSDevice* device, void* sel
                                                         12,        // Speed
                                                         PR_Now());
 
-  provider->Update(somewhere);
+  Update(somewhere);
   return;
 #else
 
-  if (!device || !device->fix)
-    return;
-
-  guint32 &fields = device->fix->fields;
-  if (!(fields & LOCATION_GPS_DEVICE_LATLONG_SET))
-    return;
+  double hErr = aDev->fix->eph/100;
+  double altitude = 0, speed = 0, track = 0;
+  if (aDev->fix->epv && !isnan(aDev->fix->epv))
+    altitude = aDev->fix->altitude;
+  if (aDev->fix->eps && !isnan(aDev->fix->eps))
+    speed = aDev->fix->speed;
+  if (aDev->fix->epd && !isnan(aDev->fix->epd))
+    track = aDev->fix->track;
 
 #ifdef LOGGING
-  fprintf (stderr, "Latitude: %.5f\nLongitude: %.5f\nAltitude: %.2f, HErr:%g\n",
-           device->fix->latitude, device->fix->longitude, device->fix->altitude, device->fix->eph/100);
-  ULOG_DEBUG_F("Latitude: %.5f\nLongitude: %.5f\nAltitude: %.2f, HErr:%g\n",
-               device->fix->latitude, device->fix->longitude, device->fix->altitude, device->fix->eph/100);
+  double dist = location_distance_between(mPrevLat, mPrevLong, aDev->fix->latitude, aDev->fix->longitude)*1000;
+  fprintf(stderr, "dist:%.9f, Lat: %.6f, Long:%.6f, HErr:%g, Alt:%.6f, VErr:%g, dir:%g[%g], sp:%g[%g]\n",
+           dist, aDev->fix->latitude, aDev->fix->longitude,
+           hErr, altitude,
+           aDev->fix->epv/2, track, aDev->fix->epd,
+           speed, aDev->fix->eps);
+  mPrevLat = aDev->fix->latitude;
+  mPrevLong = aDev->fix->longitude;
 #endif
-  nsRefPtr<nsGeoPosition> somewhere = new nsGeoPosition(device->fix->latitude,
-                                                        device->fix->longitude,
-                                                        device->fix->altitude,
-                                                        device->fix->eph/100, // Not sure about coeff
-                                                        device->fix->epv/2, // Not sure about coeff
-                                                        device->fix->track,
-                                                        device->fix->speed,
+
+  nsRefPtr<nsGeoPosition> somewhere = new nsGeoPosition(aDev->fix->latitude,
+                                                        aDev->fix->longitude,
+                                                        altitude,
+                                                        hErr,
+                                                        aDev->fix->epv/2,
+                                                        track,
+                                                        speed,
                                                         PR_Now());
-  provider->Update(somewhere);
+  Update(somewhere);
 #endif
+  return NS_OK;
 }
 
 NS_IMETHODIMP
@@ -289,7 +317,7 @@ MaemoLocationProvider::StartControl()
   NS_ENSURE_TRUE(mGPSControl, NS_ERROR_FAILURE);
 
   g_object_set(G_OBJECT(mGPSControl),
-               "preferred-interval", LOCATION_INTERVAL_DEFAULT,
+               "preferred-interval", mInterval,
                "preferred-method", LOCATION_METHOD_USER_SELECTED,
                 NULL);
   mControlError = g_signal_connect(mGPSControl, "error",
@@ -319,10 +347,47 @@ MaemoLocationProvider::StartDevice()
 
 NS_IMETHODIMP MaemoLocationProvider::Startup()
 {
-  nsresult rv;
-#ifdef TEST_BUILD_WITHOUT_DEVICE
-  rv = NS_OK;
-#else
+  nsresult rv(NS_OK);
+
+  PRInt32 freqVal = 5;
+  nsCOMPtr<nsIPrefBranch> prefs = do_GetService(NS_PREFSERVICE_CONTRACTID);
+  if (prefs) {
+    rv = prefs->GetIntPref("geo.location_update_freq", &freqVal);
+    switch (freqVal) {
+      case 0:
+        mInterval = LOCATION_INTERVAL_DEFAULT;
+        break;
+      case 1:
+        mInterval = LOCATION_INTERVAL_1S;
+        break;
+      case 2:
+        mInterval = LOCATION_INTERVAL_2S;
+        break;
+      case 5:
+        mInterval = LOCATION_INTERVAL_5S;
+        break;
+      case 10:
+        mInterval = LOCATION_INTERVAL_10S;
+        break;
+      case 20:
+        mInterval = LOCATION_INTERVAL_20S;
+        break;
+      case 30:
+        mInterval = LOCATION_INTERVAL_30S;
+        break;
+      case 60:
+        mInterval = LOCATION_INTERVAL_60S;
+        break;
+      case 120:
+        mInterval = LOCATION_INTERVAL_120S;
+        break;
+      default:
+        mInterval = LOCATION_INTERVAL_DEFAULT;
+        break;
+    }
+  }
+
+#ifndef TEST_BUILD_WITHOUT_DEVICE
   rv = StartControl();
   NS_ENSURE_SUCCESS(rv, rv);
 
@@ -333,17 +398,20 @@ NS_IMETHODIMP MaemoLocationProvider::Startup()
   if (mUpdateTimer)
     return NS_OK;
 
+  PRInt32 update = 0; //0 second no timer created
+  if (prefs)
+    prefs->GetIntPref("geo.default.update", &update);
+
+  if (!update)
+    return NS_OK;
+
   mUpdateTimer = do_CreateInstance("@mozilla.org/timer;1", &rv);
 
   if (NS_FAILED(rv))
     return NS_ERROR_FAILURE;
 
-  nsCOMPtr<nsIPrefBranch> prefs = do_GetService(NS_PREFSERVICE_CONTRACTID);
-  PRInt32 update = 3000; //3 second default timer
-  if (prefs)
-    prefs->GetIntPref("geo.default.update", &update);
-
-  mUpdateTimer->InitWithCallback(this, update, nsITimer::TYPE_REPEATING_SLACK);
+  if (update)
+    mUpdateTimer->InitWithCallback(this, update, nsITimer::TYPE_REPEATING_SLACK);
 
   return NS_OK;
 }
